#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from datetime import datetime
from uuid import UUID

from evo.common import ResourceMetadata
from evo.workspaces import ServiceUser

from .endpoints.models import BBox, BBoxXYZ, Column, RotationAxis

__all__ = ["BaseGridDefinition", "BlockModel", "RegularGridDefinition", "Version"]


@dataclass(frozen=True, kw_only=True)
class BaseGridDefinition:
    """Base class for grid definitions."""

    model_origin: list[float]
    """Origin of the block model."""
    rotations: list[tuple[RotationAxis, float]]
    """Rotation of the block model, as a list of intrinsic rotations.

    The angle is in degrees, with positive angles indicating a clockwise rotation around the axis, when looking down the
    axis towards the negative direction.
    """

    def __post_init__(self):
        if len(self.model_origin) != 3:
            raise ValueError("model_origin must have 3 elements")
        if len(self.rotations) > 3:
            raise ValueError("rotations must have less than or equal to 3 elements")


@dataclass(frozen=True, kw_only=True)
class RegularGridDefinition(BaseGridDefinition):
    """A regular grid definition, defining the grid's origin, block size, number of block, and orientation."""

    n_blocks: list[int]
    """Number of blocks in the grid along each axis."""
    block_size: list[float]
    """Size of the blocks along each axis."""

    def __post_init__(self):
        super().__post_init__()
        if len(self.n_blocks) != 3:
            raise ValueError("n_blocks must have 3 elements")
        if len(self.block_size) != 3:
            raise ValueError("block_size must have 3 elements")


@dataclass(frozen=True, kw_only=True)
class BlockModel(ResourceMetadata):
    geoscience_object_id: UUID | None
    """
    UUID of the Geoscience Object Service object associated with the block model
    """

    description: str | None
    """
    Description of the block model.
    """

    grid_definition: BaseGridDefinition
    """
    Definition of the block model grid.
    """

    coordinate_reference_system: str | None
    """
    Coordinate reference system used in the block model.
    """

    size_unit_id: str | None
    """
    Unit ID denoting the length unit used for the block model's blocks.
    """

    bbox: BBoxXYZ
    """
    Axis-aligned bounding box of the block model.

    This is the smallest box that fully contains all blocks within the block model, regardless of whether they contain data.
    It is defined by the minimum and maximum coordinates along each axis.
    """

    last_updated_at: datetime
    """
    Date and time of the last block model update, including metadata updates
    """

    last_updated_by: ServiceUser
    """
    User who last updated the block model, including metadata updates
    """

    @property
    def url(self) -> str:
        """The url of the block model version."""
        return "{hub_url}/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}".format(
            hub_url=self.environment.hub_url.rstrip("/"),
            org_id=self.environment.org_id,
            workspace_id=self.environment.workspace_id,
            bm_id=self.id,
        )


@dataclass(frozen=True, kw_only=True)
class Version:
    bm_uuid: UUID
    """UUID of the block model this version belongs to."""

    version_id: int
    """
    Identifier for the version within a block model as a monotonically increasing integer, where 1 is
    the `version_id` for the version created upon creation of the block model.
    """

    version_uuid: UUID
    """
    A universally unique identifier for the version
    """

    parent_version_id: int
    """
    Previous version. 0 if this is the first version.
    """

    base_version_id: int | None
    """
    Version the update was applied to. This will be the same as `parent_version_id`, except for
    updates made by Leapfrog, where it is the current local version when the block model is published. This is None if this
    is the first version.
    """

    geoscience_version_id: str | None
    """
    ID of the Geoscience Object Service object version associated with this block model version
    """

    created_at: datetime
    """
    User who performed the action that created the version
    """

    created_by: ServiceUser
    """
    User who performed the action that created the version
    """

    comment: str
    """
    User-supplied comment for this version
    """

    bbox: BBox | None = None
    """
    Bounding box of data updated between this version and last version. Will be None for the initial version, and updates
    that only delete and rename columns.
    """

    columns: list[Column]
    """
    Columns within this version
    """



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections.abc import Iterator
from pathlib import Path
from typing import Any
from uuid import UUID

from evo import logging
from evo.common import APIConnector, Environment, ICache, IFeedback
from evo.common.exceptions import StorageFileNotFoundError
from evo.common.io.exceptions import DataExistsError
from evo.common.utils import NoFeedback, PartialFeedback

from ..io import _CACHE_SCOPE, ObjectDataUpload

try:
    import pyarrow as pa
except ImportError:
    raise ImportError("ObjectDataClient requires the `pyarrow` package to be installed")

try:
    import pandas as pd
except ImportError:
    _PD_AVAILABLE = False
else:
    _PD_AVAILABLE = True

__all__ = ["ObjectDataClient"]

logger = logging.getLogger("object.data")

_DATA_ID_KEY = "data"  # The key used to identify data references in geoscience objects.


def _iter_refs(target: Any, _key: str | None = None) -> Iterator[str]:
    """Iterate over all data references in a geoscience object.

    :param target: Any part of a geoscience object.
    :param _key: The key of the current target, if any.

    :return: An iterator over all data references in the geoscience object.
    """
    match target:
        case dict() as nested:
            for key, value in nested.items():
                yield from _iter_refs(value, key)
        case list() as items:
            for item in items:
                yield from _iter_refs(item)
        case str() | UUID() as value if _key == _DATA_ID_KEY:
            yield str(value)


class ObjectDataClient:
    """An optional wrapper around data upload and download functionality for geoscience objects.

    This class provides a high-level interface for uploading and downloading data that is referenced in geoscience
    objects, and caching the data locally. It depends on the optional dependency `pyarrow`, which is not installed
    by default. This dependency can be installed with `pip install evo-objects[utils]`.
    """

    def __init__(self, environment: Environment, connector: APIConnector, cache: ICache) -> None:
        """
        :param environment: The environment to upload and download data from.
        :param connector: The API connector to use for uploading and downloading data.
        :param cache: The cache to use for storing data locally.
        """
        self._environment = environment
        self._connector = connector
        self._cache = cache

    @property
    def cache_location(self) -> Path:
        """The location of the cache for this client."""
        return self._cache.get_location(environment=self._environment, scope=_CACHE_SCOPE)

    def clear_cache(self) -> None:
        """Clear the cache used by this client."""
        self._cache.clear_cache(environment=self._environment, scope=_CACHE_SCOPE)

    async def upload_referenced_data(self, object_model: dict, fb: IFeedback = NoFeedback) -> None:
        """Upload all data referenced by a geoscience object.

        All referenced data that has not already been uploaded must be available in the cache. The best way of ensuring
        this is to use the `save_table` method to save the data to the cache before calling this method.

        :param object_model: The geoscience object to upload data from.
        :param fb: The feedback object to report progress to.

        :raises StorageFileNotFoundError: If the data to be uploaded is not found in the cache.
        """
        names = list(_iter_refs(object_model))
        n_names = len(names)
        logger.debug(f"Found {n_names} data reference{'' if n_names == 1 else 's'}")
        n_uploaded = 0
        async for ctx in ObjectDataUpload._create_multiple(
            connector=self._connector,
            environment=self._environment,
            names=names,
        ):
            if fb is not NoFeedback:
                fb_part = PartialFeedback(parent=fb, start=n_uploaded / n_names, end=(n_uploaded + 1) / n_names)
            else:
                fb_part = NoFeedback

            data_file = self.cache_location / ctx.name
            logger.debug(f"Uploading data: {ctx.name}")
            if not data_file.exists():
                raise StorageFileNotFoundError(f"Object data file not found in cache: {data_file}")
            await ctx.upload_from_path(filename=data_file, transport=self._connector.transport, fb=fb_part)
            n_uploaded += 1

        logger.debug(
            f"Uploaded {n_uploaded} data reference{'' if n_uploaded == 1 else 's'}, skipped {n_names - n_uploaded}"
        )
        fb.progress(1)

    def save_table(self, table: pa.Table) -> dict:
        """Save a pyarrow table to a file, returning the table info as a dictionary.

        :param table: The pyarrow table to save.

        :return: Information about the saved table.

        :raises TableFormatError: If the provided table does not match this format.
        :raises StorageFileNotFoundError: If the destination does not exist or is not a directory.
        """
        from .table_formats import get_known_format

        known_format = get_known_format(table)
        table_info = known_format.save_table(table=table, destination=self.cache_location)
        return table_info

    async def upload_table(self, table: pa.Table, fb: IFeedback = NoFeedback) -> dict:
        """Upload pyarrow table to the geoscience object service, returning a GO model of the uploaded data.

        :param table: The table to be uploaded.
        :param fb: A feedback object for tracking upload progress.

        :return: A description of the uploaded data.

        :raises TableFormatError: If the table does not match a known format.
        """
        table_info = self.save_table(table)
        upload = ObjectDataUpload(connector=self._connector, environment=self._environment, name=table_info["data"])
        try:
            await upload.upload_from_cache(cache=self._cache, transport=self._connector.transport, fb=fb)
        except DataExistsError:
            logger.debug(f"Data not uploaded because data already exists (label: {table_info['data']})")
            fb.progress(1)
        return table_info

    async def download_table(
        self, object_id: UUID, version_id: str, table_info: dict, fb: IFeedback = NoFeedback
    ) -> pa.Table:
        """Download pyarrow table from the geoscience object service.

        The parquet metadata will be used to make sure the file contents matches the expected format before the table
        is read into memory.

        :param object_id: The object ID to download the data from.
        :param version_id: The version ID to download the data from.
        :param table_info: The table info that defines the expected format. The model's `data` will be downloaded from
            the service.
        :param fb: A feedback object for tracking download progress.

        :return: A pyarrow table loaded directly from the parquet file.

        :raises DataNotFoundError: If the data does not exist or is not associated with this object version.
        :raises TableFormatError: If the data does not match the expected format.
        :raises SchemaValidationError: If the data has a different number of rows than expected.
        """
        # Import here to avoid circular import.
        from ..client import ObjectAPIClient
        from ..parquet import ParquetDownloader

        client = ObjectAPIClient(self._environment, self._connector)
        (download,) = [d async for d in client.prepare_data_download(object_id, version_id, [table_info["data"]])]

        # Defer downloading the table to the new ParquetLoader class.
        async with ParquetDownloader(
            download=download, transport=self._connector.transport, cache=self._cache
        ).with_feedback(fb) as loader:
            loader.validate_with_table_info(table_info)
            return loader.load_as_table()

    if _PD_AVAILABLE:
        # Optional support for pandas dataframes. Depends on both pyarrow and pandas.

        def save_dataframe(self, dataframe: pd.DataFrame) -> dict:
            """Save a pandas dataframe to a file, returning the table info as a dictionary.

            :param dataframe: The pandas dataframe to save.

            :return: Information about the saved table.

            :raises TableFormatError: If the provided table does not match this format.
            :raises StorageFileNotFoundError: If the destination does not exist or is not a directory.
            """
            return self.save_table(pa.Table.from_pandas(dataframe))

        async def upload_dataframe(self, dataframe: pd.DataFrame, fb: IFeedback = NoFeedback) -> dict:
            """Upload pandas dataframe to the geoscience object service, returning a GO model of the uploaded data.

            :param dataframe: The pandas dataframe to be uploaded.
            :param fb: A feedback object for tracking upload progress.

            :return: A description of the uploaded data.

            :raises TableFormatError: If the table does not match a known format.
            """
            table_info = await self.upload_table(pa.Table.from_pandas(dataframe), fb=fb)
            return table_info

        async def download_dataframe(
            self, object_id: UUID, version_id: str, table_info: dict, fb: IFeedback = NoFeedback
        ) -> pd.DataFrame:
            """Download pandas dataframe data from the geoscience object service.

            The parquet metadata will be used to make sure the file contents matches the expected format before the table
            is read into memory.

            :param object_id: The object ID to download the data from.
            :param version_id: The version ID to download the data from.
            :param table_info: The table info that defines the expected format. The model's `data` will be downloaded from
                the service.
            :param fb: A feedback object for tracking download progress.

            :return: A pandas dataframe loaded directly from the parquet file.

            :raises DataNotFoundError: If the data does not exist or is not associated with this object version.
            :raises TableFormatError: If the data does not match the expected format.
            :raises SchemaValidationError: If the data has a different number of rows than expected.
            """
            table = await self.download_table(object_id, version_id, table_info, fb)
            try:
                return table.to_pandas()
            except ModuleNotFoundError:
                raise RuntimeError("Unable to download dataframe because the `pandas` package is not installed")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from unittest import mock
from uuid import UUID

from pandas.testing import assert_frame_equal

from data import load_test_data
from evo.common import IFeedback, RequestMethod
from evo.common.io.exceptions import DataExistsError
from evo.common.test_tools import TestWithConnector, TestWithStorage
from evo.common.utils import NoFeedback, PartialFeedback
from evo.objects.utils import KnownTableFormat, ObjectDataClient
from helpers import NoImport, UnloadModule, get_sample_table_and_bytes


class TestObjectDataClient(TestWithConnector, TestWithStorage):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithStorage.setUp(self)
        self.data_client = ObjectDataClient(environment=self.environment, connector=self.connector, cache=self.cache)

    def tearDown(self) -> None:
        # Clear cache between tests to avoid cached files interfering with subsequent tests.
        self.cache.clear_cache()

    @property
    def base_path(self) -> str:
        return f"geoscience-object/orgs/{self.environment.org_id}/workspaces/{self.environment.workspace_id}"

    def test_save_table(self) -> None:
        """Test saving tabular data using pyarrow."""
        with (
            mock.patch("evo.objects.utils.table_formats.get_known_format") as mock_get_known_format,
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_table = mock.Mock()
            mock_get_known_format.return_value = mock_known_format = mock.Mock(spec=KnownTableFormat)
            mock_known_format.save_table.return_value = mock_table_info = {}

            actual_table_info = self.data_client.save_table(mock_table)

        mock_get_known_format.assert_called_once_with(mock_table)
        mock_known_format.save_table.assert_called_once_with(
            table=mock_table, destination=self.data_client.cache_location
        )
        mock_destination.upload_file.assert_not_called()
        self.transport.assert_no_requests()
        self.assertIs(mock_table_info, actual_table_info)

    def test_save_dataframe(self) -> None:
        """Test saving tabular data using pandas."""
        with (
            mock.patch("evo.objects.utils.table_formats.get_known_format") as mock_get_known_format,
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
            mock.patch("pyarrow.Table") as mock_pyarrow_table,
        ):
            mock_pyarrow_table.from_pandas.return_value = mock_table = mock.Mock()
            mock_get_known_format.return_value = mock_known_format = mock.Mock(spec=KnownTableFormat)
            mock_known_format.save_table.return_value = mock_table_info = {}

            mock_dataframe = mock.Mock()
            actual_table_info = self.data_client.save_dataframe(mock_dataframe)

        mock_get_known_format.assert_called_once_with(mock_table)
        mock_known_format.save_table.assert_called_once_with(
            table=mock_table, destination=self.data_client.cache_location
        )
        mock_destination.upload_file.assert_not_called()
        self.transport.assert_no_requests()
        self.assertIs(mock_table_info, actual_table_info)

    async def test_upload_referenced_data(self) -> None:
        put_data_response = load_test_data("put_data_batch.json")[:5]
        test_pointset = {
            "name": "Test Pointset",
            "uuid": None,
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": "unspecified",
            "locations": {
                "attributes": [
                    {
                        "name": f"Test Attribute {n}",
                        "nan_description": {"values": []},
                        "values": {
                            "data": data["name"],
                            "length": 0,
                            "width": 1,
                            "data_type": "float64",
                        },
                        "attribute_type": "scalar",
                    }
                    for n, data in enumerate(put_data_response[1:])
                ],
                "coordinates": {
                    "data": put_data_response[0]["name"],
                    "length": 0,
                    "width": 3,
                    "data_type": "float64",
                },
            },
            "schema": "/objects/pointset/1.1.0/pointset.schema.json",
        }

        data_by_name = {}
        for data in put_data_response:
            data_by_name[data["name"]] = data
            if data["exists"] is False:
                # Create an empty file to simulate the data being present in the cache.
                (self.data_client.cache_location / data["name"]).touch()

        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(put_data_response)),
            mock.patch("evo.common.io.upload.StorageDestination", autospec=True) as mock_destination,
        ):

            async def _mock_upload_file_side_effect(*args, **kwargs):
                filename = kwargs["filename"]
                data = data_by_name.pop(filename.name)
                actual_upload_url = await kwargs["url_generator"]()
                self.assertEqual(data["upload_url"], actual_upload_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIsInstance(kwargs["fb"], PartialFeedback)

            mock_destination.upload_file.side_effect = _mock_upload_file_side_effect
            await self.data_client.upload_referenced_data(test_pointset, mock.Mock(spec=IFeedback))

        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/data",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=[{"name": data["name"]} for data in put_data_response[1:]] + [{"name": put_data_response[0]["name"]}],
        )

    async def test_upload_table(self) -> None:
        """Test uploading tabular data using pyarrow or pandas."""
        put_data_response = load_test_data("put_data.json")
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(put_data_response)),
            mock.patch("evo.objects.utils.table_formats.get_known_format") as mock_get_known_format,
            mock.patch("evo.common.io.upload.StorageDestination", autospec=True) as mock_destination,
        ):
            mock_table = mock.Mock()
            mock_get_known_format.return_value = mock_known_format = mock.Mock(spec=KnownTableFormat)
            mock_known_format.save_table.return_value = mock_table_info = {}
            mock_table_info["data"] = mock_data_id = put_data_response[0]["name"]

            async def _mock_upload_file_side_effect(*args, **kwargs):
                expected_filename = self.data_client.cache_location / mock_data_id
                expected_upload_url = put_data_response[0].get("upload_url")
                actual_upload_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_upload_url, actual_upload_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

            mock_destination.upload_file.side_effect = _mock_upload_file_side_effect
            actual_table_info = await self.data_client.upload_table(mock_table)

        mock_get_known_format.assert_called_once_with(mock_table)
        mock_known_format.save_table.assert_called_once_with(
            table=mock_table, destination=self.data_client.cache_location
        )
        mock_destination.upload_file.assert_called_once()
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/data",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=[{"name": mock_data_id}],
        )
        self.assertIs(mock_table_info, actual_table_info)

    async def test_upload_dataframe(self) -> None:
        """Test uploading tabular data using pyarrow or pandas."""
        put_data_response = load_test_data("put_data.json")
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(put_data_response)),
            mock.patch("evo.objects.utils.table_formats.get_known_format") as mock_get_known_format,
            mock.patch("evo.common.io.upload.StorageDestination", autospec=True) as mock_destination,
            mock.patch("pyarrow.Table") as mock_pyarrow_table,
        ):
            mock_pyarrow_table.from_pandas.return_value = mock_table = mock.Mock()
            mock_get_known_format.return_value = mock_known_format = mock.Mock(spec=KnownTableFormat)
            mock_known_format.save_table.return_value = mock_table_info = {}
            mock_table_info["data"] = mock_data_id = put_data_response[0]["name"]

            async def _mock_upload_file_side_effect(*args, **kwargs):
                expected_filename = self.data_client.cache_location / mock_data_id
                expected_upload_url = put_data_response[0].get("upload_url")
                actual_upload_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_upload_url, actual_upload_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

            mock_destination.upload_file.side_effect = _mock_upload_file_side_effect

            mock_dataframe = mock.Mock()
            actual_table_info = await self.data_client.upload_dataframe(mock_dataframe)

        mock_get_known_format.assert_called_once_with(mock_table)
        mock_known_format.save_table.assert_called_once_with(
            table=mock_table, destination=self.data_client.cache_location
        )
        mock_destination.upload_file.assert_called_once()
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/data",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=[{"name": mock_data_id}],
        )
        self.assertIs(mock_table_info, actual_table_info)

    async def test_upload_table_exists(self) -> None:
        """Test uploading tabular data using pyarrow when the table exists."""
        put_data_response = load_test_data("put_data_exists.json")
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(put_data_response)),
            mock.patch("evo.objects.utils.table_formats.get_known_format") as mock_get_known_format,
            mock.patch("evo.common.io.upload.StorageDestination", autospec=True) as mock_destination,
        ):
            mock_table = mock.Mock()
            mock_get_known_format.return_value = mock_known_format = mock.Mock(spec=KnownTableFormat)
            mock_known_format.save_table.return_value = mock_table_info = {}
            mock_table_info["data"] = mock_data_id = put_data_response[0]["name"]

            async def _mock_upload_file_side_effect(*args, **kwargs):
                expected_filename = self.data_client.cache_location / mock_data_id
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

                with self.assertRaises(DataExistsError) as cm:
                    await kwargs["url_generator"]()
                raise cm.exception  # upload_table() should catch this exception.

            mock_destination.upload_file.side_effect = _mock_upload_file_side_effect
            actual_table_info = await self.data_client.upload_table(mock_table)

        mock_get_known_format.assert_called_once_with(mock_table)
        mock_known_format.save_table.assert_called_once_with(
            table=mock_table, destination=self.data_client.cache_location
        )
        mock_destination.upload_file.assert_called_once()
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/data",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=[{"name": mock_data_id}],
        )
        self.assertIs(mock_table_info, actual_table_info)

    async def test_upload_dataframe_exists(self) -> None:
        """Test uploading tabular data using pandas when the table exists."""
        put_data_response = load_test_data("put_data_exists.json")
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(put_data_response)),
            mock.patch("evo.objects.utils.table_formats.get_known_format") as mock_get_known_format,
            mock.patch("evo.common.io.upload.StorageDestination", autospec=True) as mock_destination,
            mock.patch("pyarrow.Table") as mock_pyarrow_table,
        ):
            mock_pyarrow_table.from_pandas.return_value = mock_table = mock.Mock()
            mock_get_known_format.return_value = mock_known_format = mock.Mock(spec=KnownTableFormat)
            mock_known_format.save_table.return_value = mock_table_info = {}
            mock_table_info["data"] = mock_data_id = put_data_response[0]["name"]

            async def _mock_upload_file_side_effect(*args, **kwargs):
                expected_filename = self.data_client.cache_location / mock_data_id
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

                with self.assertRaises(DataExistsError) as cm:
                    await kwargs["url_generator"]()
                raise cm.exception  # upload_table() should catch this exception.

            mock_destination.upload_file.side_effect = _mock_upload_file_side_effect

            mock_dataframe = mock.Mock()
            actual_table_info = await self.data_client.upload_dataframe(mock_dataframe)
            print(actual_table_info)

        mock_get_known_format.assert_called_once_with(mock_table)
        mock_known_format.save_table.assert_called_once_with(
            table=mock_table, destination=self.data_client.cache_location
        )
        mock_destination.upload_file.assert_called_once()
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/data",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=[{"name": mock_data_id}],
        )
        self.assertIs(mock_table_info, actual_table_info)

    async def test_download_table(self) -> None:
        """Test downloading tabular data using pyarrow."""
        get_object_response = load_test_data("get_object.json")
        object_id = UUID(int=2)
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)),
            mock.patch("evo.common.io.download.HTTPSource", autospec=True) as mock_source,
        ):
            mock_table_info = {
                "data": "0000000000000000000000000000000000000000000000000000000000000001",
                "length": 1,
                "width": 3,
                "data_type": "float64",
            }
            mock_data_id: str = mock_table_info["data"]
            expected_filename = self.data_client.cache_location / mock_data_id
            sample_table, payload_bytes = get_sample_table_and_bytes(
                KnownTableFormat.from_table_info(mock_table_info), 1
            )

            async def _mock_download_file_side_effect(*args, **kwargs):
                expected_download_url = get_object_response["links"]["data"][1]["download_url"]
                actual_download_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_download_url, actual_download_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])
                expected_filename.write_bytes(payload_bytes)

            mock_source.download_file.side_effect = _mock_download_file_side_effect
            actual_table = await self.data_client.download_table(object_id, None, mock_table_info)

        mock_source.download_file.assert_called_once()
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{object_id}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        self.assertEqual(sample_table, actual_table)

    async def test_download_dataframe(self) -> None:
        """Test downloading tabular data using pandas."""
        get_object_response = load_test_data("get_object.json")
        object_id = UUID(int=2)
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)),
            mock.patch("evo.common.io.download.HTTPSource", autospec=True) as mock_source,
        ):
            mock_table_info = {
                "data": "0000000000000000000000000000000000000000000000000000000000000001",
                "length": 1,
                "width": 3,
                "data_type": "float64",
            }
            mock_data_id: str = mock_table_info["data"]
            expected_filename = self.data_client.cache_location / mock_data_id
            sample_table, payload_bytes = get_sample_table_and_bytes(
                KnownTableFormat.from_table_info(mock_table_info), 1
            )

            async def _mock_download_file_side_effect(*args, **kwargs):
                expected_download_url = get_object_response["links"]["data"][1]["download_url"]
                actual_download_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_download_url, actual_download_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])
                expected_filename.write_bytes(payload_bytes)

            mock_source.download_file.side_effect = _mock_download_file_side_effect
            actual_dataframe = await self.data_client.download_dataframe(object_id, None, mock_table_info)

        mock_source.download_file.assert_called_once()
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{object_id}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        assert_frame_equal(sample_table.to_pandas(), actual_dataframe)

    async def test_download_dataframe_optional(self) -> None:
        """Test download dataframe is not available if pandas is not installed."""
        with UnloadModule("evo.objects.utils.data"), NoImport("pandas"):
            from evo.objects.utils.data import ObjectDataClient

            client = ObjectDataClient(environment=self.environment, connector=self.connector, cache=self.cache)
            self.assertFalse(
                any(
                    (
                        hasattr(ObjectDataClient, "download_dataframe"),
                        hasattr(client, "download_dataframe"),
                    )
                ),
                "download_dataframe should not be available if pandas is missing",
            )

    async def test_download_table_already_downloaded(self) -> None:
        """Test downloading tabular data using pyarrow or pandas when the table is already downloaded."""
        get_object_response = load_test_data("get_object.json")
        object_id = UUID(int=2)
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)),
            mock.patch("evo.common.io.download.HTTPSource", autospec=True) as mock_source,
        ):
            mock_table_info = {
                "data": "0000000000000000000000000000000000000000000000000000000000000001",
                "length": 1,
                "width": 3,
                "data_type": "float64",
            }
            mock_data_id: str = mock_table_info["data"]
            expected_filename = self.data_client.cache_location / mock_data_id
            sample_table, payload_bytes = get_sample_table_and_bytes(
                KnownTableFormat.from_table_info(mock_table_info), 1
            )

            async def _mock_download_file_side_effect(*args, **kwargs):
                expected_download_url = get_object_response["links"]["data"][1]["download_url"]
                actual_download_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_download_url, actual_download_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

            mock_source.download_file.side_effect = _mock_download_file_side_effect

            expected_filename.write_bytes(payload_bytes)
            actual_table = await self.data_client.download_table(object_id, None, mock_table_info)

        mock_source.download_file.assert_not_called()
        # the object metadata is still requested to get the initial download URL and check permissions.
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{object_id}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        self.transport.request.assert_called_once()  # Ensure no other requests were made.
        self.assertEqual(sample_table, actual_table)

    async def test_download_dataframe_already_downloaded(self) -> None:
        """Test downloading tabular data using pandas when the table is already downloaded."""
        get_object_response = load_test_data("get_object.json")
        object_id = UUID(int=2)
        with (
            self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)),
            mock.patch("evo.common.io.download.HTTPSource", autospec=True) as mock_source,
        ):
            mock_table_info = {
                "data": "0000000000000000000000000000000000000000000000000000000000000001",
                "length": 1,
                "width": 3,
                "data_type": "float64",
            }
            mock_data_id: str = mock_table_info["data"]
            expected_filename = self.data_client.cache_location / mock_data_id
            sample_table, payload_bytes = get_sample_table_and_bytes(
                KnownTableFormat.from_table_info(mock_table_info), 1
            )

            async def _mock_download_file_side_effect(*args, **kwargs):
                expected_download_url = get_object_response["links"]["data"][1]["download_url"]
                actual_download_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_download_url, actual_download_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

            mock_source.download_file.side_effect = _mock_download_file_side_effect

            expected_filename.write_bytes(payload_bytes)
            actual_dataframe = await self.data_client.download_dataframe(object_id, None, mock_table_info)

        mock_source.download_file.assert_not_called()
        # the object metadata is still requested to get the initial download URL and check permissions.
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{object_id}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        self.transport.request.assert_called_once()  # Ensure no other requests were made.
        assert_frame_equal(sample_table.to_pandas(), actual_dataframe)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
import logging
import time
import unittest
from unittest import mock

from evo.common.exceptions import RetryError
from evo.common.test_tools import long_test
from evo.common.utils import BackoffExponential, BackoffIncremental, BackoffLinear, Retry

logger = logging.getLogger(__name__)


class _TestException1(Exception): ...


class _TestException2(Exception): ...


class _TestException3(Exception): ...


class TestBackoffMethods(unittest.TestCase):
    def test_exponential_backoff(self) -> None:
        for factor in range(-5, 5):
            backoff = BackoffExponential(backoff_factor=factor)
            for attempt_number in range(1, 5):
                self.assertEqual(factor * (2**attempt_number), backoff.get_backoff_time(attempt_number))

    def test_max_exponential_backoff(self) -> None:
        backoff = BackoffExponential(backoff_factor=1, max_delay=4)
        self.assertEqual(2, backoff.get_backoff_time(1))
        self.assertEqual(4, backoff.get_backoff_time(10))

    def test_incremental_backoff(self) -> None:
        for factor in range(-5, 5):
            backoff = BackoffIncremental(backoff_factor=factor)
            for attempt_number in range(1, 5):
                self.assertEqual(factor * attempt_number, backoff.get_backoff_time(attempt_number))

    def test_max_incremental_backoff(self) -> None:
        backoff = BackoffIncremental(backoff_factor=1, max_delay=4)
        self.assertEqual(1, backoff.get_backoff_time(1))
        self.assertEqual(4, backoff.get_backoff_time(10))

    def test_linear_backoff(self) -> None:
        for factor in range(-5, 5):
            backoff = BackoffLinear(backoff_factor=factor)
            for attempt_number in range(1, 5):
                self.assertEqual(factor, backoff.get_backoff_time(attempt_number))

    def test_max_linear_backoff(self) -> None:
        backoff = BackoffLinear(backoff_factor=1, max_delay=4)
        self.assertEqual(1, backoff.get_backoff_time(10))

        # This is a bit of a weird case, but possible ¯\_(ツ)_/¯.
        backoff = BackoffLinear(backoff_factor=10, max_delay=4)
        self.assertEqual(4, backoff.get_backoff_time(10))


class TestRetry(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.retry = Retry(logger, max_attempts=5, backoff_method=BackoffIncremental(1))

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_successful_attempt(self, mock_sleep: mock.MagicMock) -> None:
        async for _ in self.retry:
            pass

        mock_sleep.assert_not_called()

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_max_attempts(self, mock_sleep: mock.MagicMock) -> None:
        with self.assertRaises(RetryError):
            async for handler in self.retry:
                with handler.suppress_errors():
                    raise Exception("Test exception")

        self.assertEqual(4, mock_sleep.call_count)  # 5 attempts == 4 sleeps.
        mock_sleep.assert_has_calls([mock.call(1), mock.call(2), mock.call(3), mock.call(4)])

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_reset_counter(self, mock_sleep: mock.MagicMock) -> None:
        i = 0
        with self.assertRaises(RetryError):
            async for handler in self.retry:
                if (i := i + 1) == 5:
                    handler.reset_counter()

                with handler.suppress_errors():
                    raise Exception("Test exception")

        self.assertEqual(4 + 4, mock_sleep.call_count)
        mock_sleep.assert_has_calls(
            [
                mock.call(1),
                mock.call(2),
                mock.call(3),
                mock.call(4),
                # reset_counter().
                mock.call(1),
                mock.call(2),
                mock.call(3),
                mock.call(4),
            ]
        )

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_suppress_specific_error(self, mock_sleep: mock.MagicMock) -> None:
        with self.assertRaises(RetryError):
            async for handler in self.retry:
                with handler.suppress_errors(_TestException1):
                    raise _TestException1("Expected exception")

        self.assertEqual(4, mock_sleep.call_count)  # 5 attempts == 4 sleeps.
        mock_sleep.assert_has_calls([mock.call(1), mock.call(2), mock.call(3), mock.call(4)])

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_suppress_specific_errors(self, mock_sleep: mock.MagicMock) -> None:
        errors = itertools.cycle([_TestException1("Expected exception 1"), _TestException2("Expected exception 2")])
        with self.assertRaises(RetryError):
            async for handler in self.retry:
                with handler.suppress_errors((_TestException1, _TestException2)):
                    raise next(errors)

        self.assertEqual(4, mock_sleep.call_count)  # 5 attempts == 4 sleeps.
        mock_sleep.assert_has_calls([mock.call(1), mock.call(2), mock.call(3), mock.call(4)])

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_suppress_specific_error_unexpected_not_suppressed(self, mock_sleep: mock.MagicMock) -> None:
        with self.assertRaises(_TestException2):
            async for handler in self.retry:
                with handler.suppress_errors(_TestException1):
                    raise _TestException2("Unexpected exception")

        mock_sleep.assert_not_called()

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_suppress_specific_errors_unexpected_not_suppressed(self, mock_sleep: mock.MagicMock) -> None:
        with self.assertRaises(_TestException3):
            async for handler in self.retry:
                with handler.suppress_errors((_TestException1, _TestException2)):
                    raise _TestException3("Unexpected exception")

        mock_sleep.assert_not_called()

    @mock.patch("asyncio.sleep", spec_set=True)
    async def test_handler_properties(self, mock_sleep: mock.MagicMock) -> None:
        errors = iter(
            [
                _TestException1("Attempt 1"),
                _TestException1("Attempt 2"),
                _TestException1("Attempt 3"),
                _TestException1("Attempt 4"),
                _TestException1("Attempt 5"),
            ]
        )
        with self.assertRaises(RetryError):
            async for handler in self.retry:
                # No exception suppressed yet.
                self.assertIsNone(handler.exception)
                self.assertTrue(handler.succeeded)
                self.assertFalse(handler.failed)

                this_error = next(errors)
                with handler.suppress_errors():
                    raise this_error

                # Exception suppressed.
                self.assertIs(this_error, handler.exception)
                self.assertFalse(handler.succeeded)
                self.assertTrue(handler.failed)

        self.assertEqual(4, mock_sleep.call_count)  # 5 attempts == 4 sleeps.
        mock_sleep.assert_has_calls([mock.call(1), mock.call(2), mock.call(3), mock.call(4)])

    @long_test
    async def test_actual_delay(self) -> None:
        expect_end = time.perf_counter() + 1 + 2 + 3 + 4
        with self.assertRaises(RetryError):
            async for handler in self.retry:
                with handler.suppress_errors():
                    raise Exception("Test exception")

        actual_end = time.perf_counter()
        self.assertAlmostEqual(expect_end, actual_end, delta=0.1)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from unittest import mock
from uuid import UUID

from data import load_test_data
from evo.common import Environment, HealthCheckType, Page, RequestMethod, ServiceUser
from evo.common.test_tools import BASE_URL, ORG, WORKSPACE_ID, MockResponse, TestWithConnector, utc_datetime
from evo.files import FileAPIClient, FileAPIDownload, FileAPIUpload, FileMetadata, FileVersion


class TestFileApiClient(TestWithConnector):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        self.environment = Environment(hub_url=BASE_URL, org_id=ORG.id, workspace_id=WORKSPACE_ID)
        self.file_api_client = FileAPIClient(connector=self.connector, environment=self.environment)

    @property
    def base_path(self) -> str:
        return f"file/v2/orgs/{self.environment.org_id}/workspaces/{self.environment.workspace_id}"

    async def test_check_service_health(self) -> None:
        """Test service health check implementation"""
        with mock.patch("evo.files.client.get_service_health", spec_set=True) as mock_get_service_health:
            await self.file_api_client.get_service_health()
        mock_get_service_health.assert_called_once_with(self.connector, "file", check_type=HealthCheckType.FULL)

    async def test_list_files_default_args(self) -> None:
        empty_content = load_test_data("list_files_empty.json")
        with self.transport.set_http_response(
            200, json.dumps(empty_content), headers={"Content-Type": "application/json"}
        ):
            page = await self.file_api_client.list_files()
        self.assertIsInstance(page, Page)
        self.assertEqual([], page.items())
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files?limit=5000&offset=0",
            headers={"Accept": "application/json"},
        )

    async def test_list_files_all_args(self) -> None:
        empty_content = load_test_data("list_files_empty.json")
        with self.transport.set_http_response(
            200, json.dumps(empty_content), headers={"Content-Type": "application/json"}
        ):
            page = await self.file_api_client.list_files(limit=20, offset=10, name="x.csv")
        self.assertIsInstance(page, Page)
        self.assertEqual([], page.items())
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files?limit=20&offset=10&file_name=x.csv",
            headers={"Accept": "application/json"},
        )

    async def test_list_files(self) -> None:
        content_0 = load_test_data("list_files_0.json")
        content_1 = load_test_data("list_files_1.json")
        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        page_one = await self.file_api_client.list_files(limit=2)
        expected_files_page_one = [
            FileMetadata(
                environment=self.environment,
                id=UUID(int=2),
                name="x.csv",
                created_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                created_at=utc_datetime(2020, 1, 1, 1, 30),
                modified_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=utc_datetime(2020, 1, 1, 1, 30),
                parent="/A/",
                size=11,
                version_id="2",
            ),
            FileMetadata(
                environment=self.environment,
                id=UUID(int=3),
                name="y.csv",
                created_by=ServiceUser(
                    id=UUID(int=17),
                    name=None,
                    email=None,
                ),
                created_at=utc_datetime(2020, 1, 2, 1, 30),
                modified_by=ServiceUser(
                    id=UUID(int=17),
                    name=None,
                    email=None,
                ),
                modified_at=utc_datetime(2020, 1, 2, 1, 30),
                parent="/A/",
                size=12,
                version_id="1",
            ),
        ]
        self.assertIsInstance(page_one, Page)
        self.assertEqual(expected_files_page_one, page_one.items())
        self.assertEqual(0, page_one.offset)
        self.assertEqual(2, page_one.limit)
        self.assertFalse(page_one.is_last)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files?limit=2&offset=0",
            headers={"Accept": "application/json"},
        )
        self.transport.request.reset_mock()

        page_two = await self.file_api_client.list_files(offset=page_one.next_offset, limit=page_one.limit)
        expected_files_page_two = [
            FileMetadata(
                environment=self.environment,
                id=UUID(int=4),
                name="z.csv",
                created_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                created_at=utc_datetime(2020, 1, 1, 1, 30),
                modified_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=utc_datetime(2020, 1, 1, 1, 30),
                parent="/B/",
                size=13,
                version_id="3",
            ),
        ]
        self.assertIsInstance(page_two, Page)
        self.assertEqual(expected_files_page_two, page_two.items())
        self.assertEqual(2, page_two.offset)
        self.assertEqual(2, page_two.limit)
        self.assertTrue(page_two.is_last)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files?limit=2&offset=2",
            headers={"Accept": "application/json"},
        )

    async def test_list_all_files(self) -> None:
        content_0 = load_test_data("list_files_0.json")
        content_1 = load_test_data("list_files_1.json")
        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        file_list = await self.file_api_client.list_all_files(limit_per_request=2)
        expected_files = [
            FileMetadata(
                environment=self.environment,
                id=UUID(int=2),
                name="x.csv",
                created_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                created_at=utc_datetime(2020, 1, 1, 1, 30),
                modified_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=utc_datetime(2020, 1, 1, 1, 30),
                parent="/A/",
                size=11,
                version_id="2",
            ),
            FileMetadata(
                environment=self.environment,
                id=UUID(int=3),
                name="y.csv",
                created_by=ServiceUser(
                    id=UUID(int=17),
                    name=None,
                    email=None,
                ),
                created_at=utc_datetime(2020, 1, 2, 1, 30),
                modified_by=ServiceUser(
                    id=UUID(int=17),
                    name=None,
                    email=None,
                ),
                modified_at=utc_datetime(2020, 1, 2, 1, 30),
                parent="/A/",
                size=12,
                version_id="1",
            ),
            FileMetadata(
                environment=self.environment,
                id=UUID(int=4),
                name="z.csv",
                created_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                created_at=utc_datetime(2020, 1, 1, 1, 30),
                modified_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=utc_datetime(2020, 1, 1, 1, 30),
                parent="/B/",
                size=13,
                version_id="3",
            ),
        ]

        self.assertEqual(expected_files, file_list)
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files?limit=2&offset=0",
            headers={"Accept": "application/json"},
        )
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files?limit=2&offset=2",
            headers={"Accept": "application/json"},
        )

    async def test_get_file_by_path(self) -> None:
        get_file_response = load_test_data("get_file.json")
        path = "points.csv"
        expected_metadata = FileMetadata(
            environment=self.environment,
            id=UUID(int=6),
            name="points.csv",
            created_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            created_at=utc_datetime(2020, 1, 1, 1, 30),
            modified_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            modified_at=utc_datetime(2020, 1, 1, 1, 30),
            parent="/",
            size=10,
            version_id="1",
        )

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_file_response), headers={"Content-Type": "application/json"}
        ):
            actual_metadata = await self.file_api_client.get_file_by_path(path)

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files/path/{path}",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(expected_metadata, actual_metadata)

    async def test_prepare_download_by_path(self) -> None:
        get_file_response = load_test_data("get_file.json")
        path = "points.csv"
        expected_metadata = FileMetadata(
            environment=self.environment,
            id=UUID(int=6),
            name="points.csv",
            created_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            created_at=utc_datetime(2020, 1, 1, 1, 30),
            modified_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            modified_at=utc_datetime(2020, 1, 1, 1, 30),
            parent="/",
            size=10,
            version_id="1",
        )

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_file_response), headers={"Content-Type": "application/json"}
        ):
            download = await self.file_api_client.prepare_download_by_path(path)

        self.assertIsInstance(download, FileAPIDownload)
        self.assertEqual(expected_metadata, download.metadata)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files/path/{path}",
            headers={"Accept": "application/json"},
        )

    async def test_get_file_by_id(self) -> None:
        get_file_response = load_test_data("get_file.json")
        file_id = UUID(int=6)
        expected_metadata = FileMetadata(
            environment=self.environment,
            id=file_id,
            name="points.csv",
            created_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            created_at=utc_datetime(2020, 1, 1, 1, 30),
            modified_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            modified_at=utc_datetime(2020, 1, 1, 1, 30),
            parent="/",
            size=10,
            version_id="1",
        )

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_file_response), headers={"Content-Type": "application/json"}
        ):
            actual_metadata = await self.file_api_client.get_file_by_id(file_id)

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files/{file_id}",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(expected_metadata, actual_metadata)

    async def test_prepare_download_by_id(self) -> None:
        get_file_response = load_test_data("get_file.json")
        file_id = UUID(int=6)
        expected_metadata = FileMetadata(
            environment=self.environment,
            id=file_id,
            name="points.csv",
            created_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            created_at=utc_datetime(2020, 1, 1, 1, 30),
            modified_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            modified_at=utc_datetime(2020, 1, 1, 1, 30),
            parent="/",
            size=10,
            version_id="1",
        )

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_file_response), headers={"Content-Type": "application/json"}
        ):
            download = await self.file_api_client.prepare_download_by_id(file_id)

        self.assertIsInstance(download, FileAPIDownload)
        self.assertEqual(expected_metadata, download.metadata)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files/{file_id}",
            headers={"Accept": "application/json"},
        )

    async def test_list_versions_by_path(self) -> None:
        file_path = "points.csv"
        list_versions_response = load_test_data("list_versions.json")
        with self.transport.set_http_response(
            200, json.dumps(list_versions_response), headers={"Content-Type": "application/json"}
        ):
            versions = await self.file_api_client.list_versions_by_path(file_path)
        expected_versions = [
            FileVersion(
                version_id="3",
                created_at=utc_datetime(2020, 1, 3, 1, 30),
                created_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
            ),
            FileVersion(
                version_id="2",
                created_at=utc_datetime(2020, 1, 2, 1, 30),
                created_by=ServiceUser(
                    id=UUID(int=17),
                    name="x z",
                    email="test@example.com",
                ),
            ),
            FileVersion(
                version_id="1",
                created_at=utc_datetime(2020, 1, 1, 1, 30),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000012"),
                    name="x w",
                    email="test@example.com",
                ),
            ),
        ]

        self.assertEqual(versions, expected_versions)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files/path/points.csv?include_versions=True",
            headers={"Accept": "application/json"},
        )

    async def test_list_versions_by_id(self) -> None:
        file_id = UUID(int=6)
        list_versions_response = load_test_data("list_versions.json")
        with self.transport.set_http_response(
            200, json.dumps(list_versions_response), headers={"Content-Type": "application/json"}
        ):
            versions = await self.file_api_client.list_versions_by_id(file_id)
        expected_versions = [
            FileVersion(
                version_id="3",
                created_at=utc_datetime(2020, 1, 3, 1, 30),
                created_by=ServiceUser(
                    id=UUID(int=16),
                    name="x y",
                    email="test@example.com",
                ),
            ),
            FileVersion(
                version_id="2",
                created_at=utc_datetime(2020, 1, 2, 1, 30),
                created_by=ServiceUser(
                    id=UUID(int=17),
                    name="x z",
                    email="test@example.com",
                ),
            ),
            FileVersion(
                version_id="1",
                created_at=utc_datetime(2020, 1, 1, 1, 30),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000012"),
                    name="x w",
                    email="test@example.com",
                ),
            ),
        ]

        self.assertEqual(versions, expected_versions)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/files/{file_id}?include_versions=True",
            headers={"Accept": "application/json"},
        )

    async def test_prepare_upload_by_path(self) -> None:
        upsert_file_response = load_test_data("upsert_file.json")
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(upsert_file_response), headers={"Content-Type": "application/json"}
        ):
            upload = await self.file_api_client.prepare_upload_by_path("points.csv")

        self.assertIsInstance(upload, FileAPIUpload)
        self.assertEqual(upsert_file_response["file_id"], str(upload.file_id))
        self.assertEqual(upsert_file_response["version_id"], upload.version_id)
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/files/path/points.csv",
            headers={"Accept": "application/json"},
        )

    async def test_update_file_by_id(self) -> None:
        update_file_response = load_test_data("update_file.json")
        file_id = UUID(int=5)
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(update_file_response), headers={"Content-Type": "application/json"}
        ):
            upload = await self.file_api_client.prepare_upload_by_id(file_id)

        self.assertIsInstance(upload, FileAPIUpload)
        self.assertEqual(update_file_response["file_id"], str(upload.file_id))
        self.assertEqual(update_file_response["version_id"], upload.version_id)
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/files/{file_id}",
            headers={"Accept": "application/json"},
        )

    async def test_delete_file_by_path(self) -> None:
        path = "points.csv"
        with self.transport.set_http_response(
            status_code=204, content="", headers={"Content-Type": "application/json"}
        ):
            await self.file_api_client.delete_file_by_path(path)
        self.assert_request_made(
            method=RequestMethod.DELETE,
            path=f"{self.base_path}/files/path/{path}",
        )

    async def test_delete_file_by_id(self) -> None:
        file_id = UUID(int=6)
        with self.transport.set_http_response(
            status_code=204, content="", headers={"Content-Type": "application/json"}
        ):
            await self.file_api_client.delete_file_by_id(file_id)
        self.assert_request_made(
            method=RequestMethod.DELETE,
            path=f"{self.base_path}/files/{file_id}",
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Iterator, Sequence
from pathlib import Path
from uuid import UUID

from evo import logging
from evo.common import APIConnector, Environment, ICache, IFeedback, ITransport
from evo.common.io import Download, Upload
from evo.common.io.exceptions import DataExistsError, DataNotFoundError, RenewalError
from evo.common.utils import NoFeedback, Retry

from .data import ObjectMetadata
from .endpoints import DataApi, ObjectsApi
from .endpoints.models import DataUploadRequestBody, DataUploadResponseBody

__all__ = [
    "ObjectDataDownload",
    "ObjectDataUpload",
]

logger = logging.getLogger("object.data")

_CACHE_SCOPE = "geoscience-object"
_MAX_UPLOAD_URLS = 32


class ObjectDataUpload(Upload):
    """Context for uploading data to the Geoscience Object Service."""

    def __init__(
        self, connector: APIConnector, environment: Environment, name: str, initial_url: str | None = None
    ) -> None:
        """
        :param connector: The API connector to use.
        :param environment: The environment to upload the data to.
        :param name: The name of the data to be uploaded.
        :param initial_url: An initial URL to use for the upload.
        """
        self._api = DataApi(connector)
        self._environment = environment
        self._name = name
        self._mutex = asyncio.Lock()
        self._initial_url = initial_url

    @property
    def name(self) -> str:
        """The name of the data to be uploaded."""
        return self._name

    @property
    def label(self) -> str:
        return self._name

    @property
    def environment(self) -> Environment:
        """The environment that the data will be uploaded to."""
        return self._environment

    async def get_upload_url(self) -> str:
        # Use the initial URL.
        async with self._mutex:
            if (url := self._initial_url) is not None:
                # Clear the initial URL so it can't be reused.
                self._initial_url = None
                return url

        # Get a new upload URL.
        logger.debug(f"Requesting upload URL for {self.label}")
        match await self._api.put_data_in_workspace(
            org_id=str(self.environment.org_id),
            workspace_id=str(self.environment.workspace_id),
            data_upload_request_body=[DataUploadRequestBody(name=self._name)],
        ):
            case [DataUploadResponseBody(upload_url=url)] if isinstance(url, str):
                return url
            case [DataUploadResponseBody(exists=exists)] if exists is True:
                raise DataExistsError(
                    "Named data already exists in the workspace:\n"
                    f"data_name: {self._name}\n"
                    f"workspace_id: {self.environment.workspace_id}"
                )
            case _:
                raise RenewalError("Failed to get an upload url")

    async def upload_from_cache(
        self,
        cache: ICache,
        transport: ITransport,
        max_workers: int | None = None,
        retry: Retry | None = None,
        fb: IFeedback = NoFeedback,
    ) -> None:
        target = cache.get_location(environment=self.environment, scope=_CACHE_SCOPE) / self.name
        await self.upload_from_path(filename=target, transport=transport, max_workers=max_workers, retry=retry, fb=fb)

    @staticmethod
    async def _create_multiple(
        connector: APIConnector, environment: Environment, names: Sequence[str | UUID]
    ) -> AsyncIterator[ObjectDataUpload]:
        """Internal method to create multiple data upload contexts.

        Object data that already exists in the workspace will be skipped.

        :param connector: The API connector to use.
        :param environment: The environment to upload the data to.
        :param names: The names of the data to be uploaded.

        :return: An asynchronous iterator of data upload contexts.
        """
        api = DataApi(connector)
        # Up to _MAX_UPLOAD_URLS upload URLs can be generated in a single API request. We can accommodate any number of
        # data names by batching them in groups of _MAX_UPLOAD_URLS.
        for start in range(0, len(names), _MAX_UPLOAD_URLS):
            end = min(start + _MAX_UPLOAD_URLS, len(names))
            logger.debug(f"Requesting {end - start} upload URL{'' if end - start == 1 else 's'}")
            response = await api.put_data_in_workspace(
                org_id=str(environment.org_id),
                workspace_id=str(environment.workspace_id),
                data_upload_request_body=[DataUploadRequestBody(name=str(name)) for name in names[start:end]],
            )
            for ref in response:
                if ref.exists:
                    logger.debug(f"Skipping upload of existing data: {ref.name}")
                    continue
                yield ObjectDataUpload(
                    connector=connector,
                    environment=environment,
                    name=ref.name,
                    initial_url=ref.upload_url,
                )


class ObjectDataDownload(Download[ObjectMetadata]):
    """Context for downloading data from the Geoscience Object Service."""

    def __init__(
        self, connector: APIConnector, metadata: ObjectMetadata, name: str, initial_url: str | None = None
    ) -> None:
        self._api = ObjectsApi(connector)
        self._metadata = metadata
        self._name = name
        self._mutex = asyncio.Lock()
        self._initial_url = initial_url

    @property
    def name(self) -> str:
        """The name of the data to be uploaded."""
        return self._name

    @property
    def label(self) -> str:
        return f"{self.metadata.path} (ref={self.name})"

    @property
    def metadata(self) -> ObjectMetadata:
        return self._metadata

    def _get_cache_location(self, cache: ICache) -> Path:
        return cache.get_location(environment=self.metadata.environment, scope=_CACHE_SCOPE) / self._name

    async def get_download_url(self) -> str:
        # Use the initial URL.
        async with self._mutex:
            if (url := self._initial_url) is not None:
                # Clear the initial URL so it can't be reused.
                self._initial_url = None
                return url

        # Get a new download URL.
        logger.debug(f"Requesting download URL for {self.label}")
        response = await self._api.get_object_by_id(
            org_id=str(self.metadata.environment.org_id),
            workspace_id=str(self.metadata.environment.workspace_id),
            object_id=str(self.metadata.id),
            version=self.metadata.version_id,
        )
        for data_ref in response.links.data:
            data_key = getattr(data_ref, "name", data_ref.id)
            if data_key != self._name:
                continue
            elif isinstance(url := data_ref.download_url, str):
                return url
            else:
                raise RenewalError(f"Failed to get a data download url for {self.label}")
        raise DataNotFoundError(f"Unable to find the requested data: {self.label}")

    @staticmethod
    def _create_multiple(
        connector: APIConnector, metadata: ObjectMetadata, urls_by_name: dict[str, str]
    ) -> Iterator[ObjectDataDownload]:
        """Internal method to create multiple data download contexts.

        :param connector: The API connector to use.
        :param metadata: The metadata of the object to download data from.
        :param urls_by_name: A mapping of data names to download URLs.

        :return: An asynchronous iterator of data download contexts.
        """
        for name, url in urls_by_name.items():
            yield ObjectDataDownload(
                connector=connector,
                metadata=metadata,
                name=name,
                initial_url=url,
            )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import contextlib
from collections.abc import AsyncGenerator
from io import BytesIO
from logging import getLogger
from types import TracebackType
from typing import cast

import pyarrow as pa
import pyarrow.parquet as pq
from pydantic import TypeAdapter

from evo.common import ICache, IFeedback, ITransport
from evo.common.io import BytesDestination, ChunkedIOManager, Download, HTTPSource
from evo.common.utils import NoFeedback

from ..exceptions import SchemaValidationError, TableFormatError
from ..utils import ArrowTableFormat, KnownTableFormat
from .types import TableInfo

try:
    import pandas as pd
except ImportError:
    _PD_AVAILABLE = False
else:
    _PD_AVAILABLE = True

try:
    import numpy as np
except ImportError:
    _NP_AVAILABLE = False
else:
    _NP_AVAILABLE = True

__all__ = [
    "ParquetDownloader",
    "ParquetLoader",
]

logger = getLogger(__name__)

_TABLE_INFO_ADAPTER: TypeAdapter[TableInfo] = TypeAdapter(TableInfo)


class ParquetLoader:
    """A loader for Parquet data from a pyarrow.parquet.ParquetFile.

    This class adds standardised support for validating Geoscience Object table info
    against the loaded Parquet schema, as well as convenience methods for loading
    the data as a PyArrow Table, Pandas DataFrame, or NumPy array.
    """

    def __init__(self, pa_file: pa.NativeFile) -> None:
        """
        :param pa_file: A PyArrow NativeFile containing the Parquet data.
        """
        self._pa_file = pa_file
        self._parquet_file: pq.ParquetFile | None = None

    def __enter__(self) -> ParquetLoader:
        if self._parquet_file is not None:
            raise RuntimeError("ParquetLoader is already in use")
        self._parquet_file = pq.ParquetFile(self._pa_file.__enter__())
        return self

    async def __aenter__(self) -> ParquetLoader:
        # Delegate to the synchronous context manager.
        # This implementation is just to support async with
        # syntax for combination with ParquetDownloader below.
        return self.__enter__()

    def __exit__(
        self,
        exc_type: type[Exception] | None,
        exc_val: Exception | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._parquet_file = None
        return self._pa_file.__exit__(exc_type, exc_val, exc_tb)

    async def __aexit__(
        self,
        exc_type: type[Exception] | None,
        exc_val: Exception | None,
        exc_tb: TracebackType | None,
    ) -> None:
        # Delegate to the synchronous context manager.
        # This implementation is just to support async with
        # syntax for combination with ParquetDownloader below.
        return self.__exit__(exc_type, exc_val, exc_tb)

    def validate_with_table_info(self, table_info: TableInfo) -> None:
        """Validate the provided TableInfo against the loaded Parquet schema.

        :param table_info: The TableInfo to validate against the loaded Parquet schema.

        :raises SchemaValidationError: If the loaded Parquet schema does not match the expected schema.
        """
        if (pa_file := self._parquet_file) is None:
            raise RuntimeError("ParquetLoader context is not active")

        logger.debug("Checking parquet data format")

        validated_table_info = _TABLE_INFO_ADAPTER.validate_python(table_info)
        expected_format = KnownTableFormat.from_table_info(validated_table_info)
        actual_format = ArrowTableFormat.from_schema(pa_file.schema_arrow)
        try:
            expected_format._check_format(actual_format)
        except TableFormatError as e:
            raise SchemaValidationError(str(e)) from None

        logger.debug("Checking parquet data length")
        actual_length = pa_file.metadata.num_rows
        if table_info["length"] != actual_length:
            raise SchemaValidationError(
                f"Row count ({actual_length}) does not match expectation ({table_info['length']})"
            )

        logger.debug("Parquet metadata checks succeeded")

    def load_as_table(self) -> pa.Table:
        """Load the Parquet data as a PyArrow Table."""
        if self._parquet_file is None:
            raise RuntimeError("ParquetLoader context is not active")
        else:
            return self._parquet_file.read()

    if _PD_AVAILABLE:
        # Optional support for pandas dataframes

        def load_as_dataframe(self) -> pd.DataFrame:
            """Load the Parquet data as a Pandas DataFrame."""
            table = self.load_as_table()
            return table.to_pandas()

    if _NP_AVAILABLE:
        # Optional support for numpy arrays

        def load_as_array(self) -> np.ndarray:
            """Load the Parquet data as a NumPy array.

            The array will have a shape of (N,) for single-column data or (N, M) for multi-column data,
            where N is the number of rows and M is the number of columns. The target data _must_ have a uniform dtype.

            :return: A NumPy array containing the data.
            """
            table = self.load_as_table()
            columns = cast(
                list[np.ndarray], [col.combine_chunks().to_numpy(zero_copy_only=False) for col in table.itercolumns()]
            )
            if len(columns) == 1:
                return columns[0]
            else:
                return np.column_stack(columns)


class ParquetDownloader:
    """A downloader for Parquet data that provides a ParquetLoader for reading the data.

    This class supports downloading the data to a cache or to memory, and provides
    a ParquetLoader for reading the downloaded data.
    """

    def __init__(self, download: Download, transport: ITransport, cache: ICache | None = None) -> None:
        """
        :param download: The download information for the Parquet data.
        :param transport: The transport to use for data downloads.
        :param cache: An optional cache to use for data downloads.
        """
        self._evo_download = download
        self._transport = transport
        self._cache = cache

    async def _download_to_cache(self, fb: IFeedback) -> pa.OSFile:
        cached = await self._evo_download.download_to_cache(self._cache, self._transport, fb=fb)
        return pa.OSFile(str(cached), "r")

    async def _download_to_memory(self, fb: IFeedback) -> pa.BufferReader:
        # Initialize a buffer to store the downloaded data in memory
        memory = BytesIO()

        # Use ChunkedIOManager to download the data into the memory buffer
        manager = ChunkedIOManager()
        async with HTTPSource(self._evo_download.get_download_url, self._transport) as source:
            destination = BytesDestination(memory)
            await manager.run(source, destination, fb=fb)

        # Reset the buffer's position to the beginning
        memory.seek(0)
        return pa.BufferReader(memory.getbuffer())

    async def download(self, fb: IFeedback = NoFeedback) -> ParquetLoader:
        """Download the Parquet data and return a ParquetLoader for reading it.

        :param fb: An optional feedback instance to report download progress to.

        :return: A ParquetLoader that can be used to read the downloaded data.
        """
        if self._cache is not None:
            file = await self._download_to_cache(fb)
        else:
            file = await self._download_to_memory(fb)

        return ParquetLoader(file)

    @contextlib.asynccontextmanager
    async def __aenter__(self) -> AsyncGenerator[ParquetLoader, None]:
        # Delegate to the download method to get a ParquetLoader.
        async with await self.download() as loader:
            yield loader

    @contextlib.asynccontextmanager
    async def with_feedback(self, fb: IFeedback) -> AsyncGenerator[ParquetLoader, None]:
        """Async context manager to download the Parquet data with feedback and provide a ParquetLoader for reading it.

        :param fb: A feedback instance to report download progress to.

        :yields: A ParquetLoader that can be used to read the downloaded data.
        """
        async with await self.download(fb=fb) as loader:
            yield loader



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Colormap API
=============


        The Evo Colormap API allows you to create and manage colormaps, and associate them to attributes.


This code is generated from the OpenAPI specification for Colormap API.
API version: 1.0
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod  # noqa: F401

from ..models import *  # noqa: F403

__all__ = ["AssociationsApi"]


class AssociationsApi:
    """Api client for the Associations endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def get_association_by_id(
        self,
        org_id: str,
        workspace_id: str,
        object_id: str,
        association_id: str,
        api_preview: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> AttributeColormapAssociation:  # noqa: F405
        """Get attribute association by ID

        Returns a specified object attribute association by its ID, for a specified workspace and organization.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param object_id:
            Format: `uuid`
            Example: `'object_id_example'`
        :param association_id:
            Format: `uuid`
            Example: `'association_id_example'`
        :param api_preview: (optional)
            Example: `'api_preview_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseRFC87Error: If the server responds with any other HTTP status between
            400 and 599, and the body of the response conforms to RFC 87.
        :raise evo.common.exceptions.EvoApiException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not conform to RFC 87.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "object_id": object_id,
            "association_id": association_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if api_preview is not None:
            _header_params["api-preview"] = api_preview
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": AttributeColormapAssociation,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/colormap/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}/associations/{association_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_association_collection(
        self,
        org_id: str,
        workspace_id: str,
        object_id: str,
        api_preview: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> AttributeColormapAssociationList:  # noqa: F405
        """List object attribute associations

        Get a list of all object attribute associations for a specified workspace and organization.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param object_id:
            Format: `uuid`
            Example: `'object_id_example'`
        :param api_preview: (optional)
            Example: `'api_preview_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseRFC87Error: If the server responds with any other HTTP status between
            400 and 599, and the body of the response conforms to RFC 87.
        :raise evo.common.exceptions.EvoApiException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not conform to RFC 87.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "object_id": object_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if api_preview is not None:
            _header_params["api-preview"] = api_preview
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": AttributeColormapAssociationList,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/colormap/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}/associations",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def post_association(
        self,
        org_id: str,
        workspace_id: str,
        object_id: str,
        attribute_association_data: AttributeAssociationData,  # noqa: F405
        api_preview: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> AttributeColormapAssociation:  # noqa: F405
        """Create object attribute association

        Create a new association for a colormap to an object attribute.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param object_id:
            Format: `uuid`
            Example: `'object_id_example'`
        :param attribute_association_data:
            Example: `endpoints.AttributeAssociationData()`
        :param api_preview: (optional)
            Example: `'api_preview_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseRFC87Error: If the server responds with any other HTTP status between
            400 and 599, and the body of the response conforms to RFC 87.
        :raise evo.common.exceptions.EvoApiException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not conform to RFC 87.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "object_id": object_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if api_preview is not None:
            _header_params["api-preview"] = api_preview
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": AttributeColormapAssociation,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/colormap/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}/associations",
            path_params=_path_params,
            header_params=_header_params,
            body=attribute_association_data,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def post_batch_associations(
        self,
        org_id: str,
        workspace_id: str,
        object_id: str,
        attribute_associations_data: AttributeAssociationsData,  # noqa: F405
        api_preview: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> AttributeColormapAssociationList:  # noqa: F405
        """Create multiple object attribute associations

        Create new associations for a colormap to multiple object attributes.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param object_id:
            Format: `uuid`
            Example: `'object_id_example'`
        :param attribute_associations_data:
            Example: `endpoints.AttributeAssociationsData()`
        :param api_preview: (optional)
            Example: `'api_preview_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseRFC87Error: If the server responds with any other HTTP status between
            400 and 599, and the body of the response conforms to RFC 87.
        :raise evo.common.exceptions.EvoApiException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not conform to RFC 87.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "object_id": object_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if api_preview is not None:
            _header_params["api-preview"] = api_preview
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": AttributeColormapAssociationList,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/colormap/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}/associations/batch",
            path_params=_path_params,
            header_params=_header_params,
            body=attribute_associations_data,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from typing import Any

from parameterized import parameterized

from evo.common import DependencyStatus, HealthCheckType, RequestMethod, ServiceHealth, ServiceStatus
from evo.common.exceptions import ServiceHealthCheckFailed
from evo.common.test_tools import TestWithConnector
from evo.common.utils import get_service_health, get_service_status

from ...data import load_test_data


def using_test_data(full: bool, with_dependencies: bool, strict: bool) -> Any:
    test_data = load_test_data("health_check.json")
    p_input = []
    for scenario in test_data:
        if full:
            content = {
                "status": scenario["status"],
                "version": scenario["version"],
            }
            if with_dependencies and "dependencies" in scenario:
                content["dependencies"] = scenario["dependencies"]
        else:
            content = scenario["status"]

        if strict:
            status_code = 200 if scenario["status"] == "pass" else 503
        else:
            status_code = 200 if scenario["status"] in {"pass", "degraded"} else 503

        p_input.append((scenario["version"], content, status_code))
    return parameterized.expand(p_input)


class TestHealthCheck(TestWithConnector):
    @using_test_data(full=False, with_dependencies=False, strict=False)
    async def test_get_service_status(self, _label: str, content: str, status_code: int) -> None:
        with self.transport.set_http_response(status_code, content):
            status = await get_service_status(self.connector, "test", check_type=HealthCheckType.BASIC)

        self.assert_request_made(RequestMethod.GET, "/test/health_check")
        self.assertEqual(ServiceStatus(content), status)

    @using_test_data(full=False, with_dependencies=True, strict=False)
    async def test_get_service_status_with_dependencies(self, _label: str, content: str, status_code: int) -> None:
        with self.transport.set_http_response(status_code, content):
            status = await get_service_status(self.connector, "test", check_type=HealthCheckType.FULL)

        self.assert_request_made(RequestMethod.GET, "/test/health_check?check_dependencies=true")
        self.assertEqual(ServiceStatus(content), status)

    @using_test_data(full=False, with_dependencies=True, strict=True)
    async def test_get_service_status_with_dependencies_strict(
        self, _label: str, content: str, status_code: int
    ) -> None:
        with self.transport.set_http_response(status_code, content):
            status = await get_service_status(self.connector, "test", check_type=HealthCheckType.STRICT)

        self.assert_request_made(RequestMethod.GET, "/test/health_check?check_dependencies=true&strict=true")
        self.assertEqual(ServiceStatus(content), status)

    def _check_health_check_response(self, content: dict, status_code: int, response: ServiceHealth) -> None:
        self.assertEqual("test", response.service)
        self.assertEqual(status_code, response.status_code)
        self.assertEqual(ServiceStatus(content["status"]), response.status)
        self.assertEqual(content["version"], response.version)

        if "dependencies" in content:
            self.assertIsInstance(response.dependencies, dict)
            for dependency_name, dependency_status in content["dependencies"].items():
                self.assertIn(dependency_name, response.dependencies)
                self.assertEqual(DependencyStatus(dependency_status), response.dependencies[dependency_name])

        if status_code == 200:
            response.raise_for_status()
        else:
            with self.assertRaises(ServiceHealthCheckFailed):
                response.raise_for_status()

    @using_test_data(full=True, with_dependencies=False, strict=False)
    async def test_get_service_health(self, _label: str, content: dict, status_code: int) -> None:
        with self.transport.set_http_response(status_code, json.dumps(content)):
            response = await get_service_health(self.connector, "test", check_type=HealthCheckType.BASIC)

        self.assert_request_made(RequestMethod.GET, "/test/health_check?full=true")
        self._check_health_check_response(content, status_code, response)

    @using_test_data(full=True, with_dependencies=True, strict=False)
    async def test_get_service_health_with_dependencies(self, _label: str, content: dict, status_code: int) -> None:
        with self.transport.set_http_response(status_code, json.dumps(content)):
            response = await get_service_health(self.connector, "test", check_type=HealthCheckType.FULL)

        self.assert_request_made(RequestMethod.GET, "/test/health_check?full=true&check_dependencies=true")
        self._check_health_check_response(content, status_code, response)

    @using_test_data(full=True, with_dependencies=True, strict=True)
    async def test_get_service_health_with_dependencies_strict(
        self, _label: str, content: dict, status_code: int
    ) -> None:
        with self.transport.set_http_response(status_code, json.dumps(content)):
            response = await get_service_health(self.connector, "test", check_type=HealthCheckType.STRICT)

        self.assert_request_made(RequestMethod.GET, "/test/health_check?full=true&check_dependencies=true&strict=true")
        self._check_health_check_response(content, status_code, response)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Block Model API
=============


    The Block Model API provides the ability to manage and report on block models in your Evo workspaces. Enable your
    product with Evo connected workflows by integrating with the Seequent Block Model API.

    The Block Model API supports a range of sub-blocking options, and both full and partial updates to block models,
    including updates of specified columns and/or sub-volumes. Block models are versioned, and the service supports
    reporting on the material content of current or previous versions, and comparing the content between versions of
    a block model.

    For more information on the Block Model API, see [Overview](/docs/blockmodel/overview), or the API references here.


This code is generated from the OpenAPI specification for Block Model API.
API version: 1.16.1
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod  # noqa: F401

from ..models import *  # noqa: F403

__all__ = ["ReportsApi"]


class ReportsApi:
    """API client for the Reports endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def create_report_specification(
        self,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        create_report_specification: CreateReportSpecification,  # noqa: F405
        run_now: bool | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportSpecificationWithJobUrl:  # noqa: F405
        """Create a Report Specification

        Create a new Report Specification for the block model.  The Report Specification defines the columns and categories that will be used to generate reports for the block model. The cut-off columns are used to determine whether a block is included in the report. A list of values can be provided in `cut_off_values`, each of which will be calculated as a sub-report.  To calculate column totals, density information is derived from either the block model data in the column indicated by `density_col_id`, or if `density_col_id` is unset or contains null values, the `density_value` is used.

        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param create_report_specification:
            Example: `endpoints.CreateReportSpecification()`
        :param run_now: (optional) Whether to trigger a Reporting Job immediately after the creation of the Report Specification
            Example: `False`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if run_now is not None:
            _query_params["run_now"] = run_now

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": ReportSpecificationWithJobUrl,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=create_report_specification,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_report_result(
        self,
        rs_id: str,
        report_result_uuid: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportResult:  # noqa: F405
        """Get a Report Result

        Get a Report Result

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param report_result_uuid: ID of the Report Result that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "report_result_uuid": report_result_uuid,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ReportResult,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}/results/{report_result_uuid}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_report_result_comparison(
        self,
        rs_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        var_from: str,
        to: str,
        decimal_places: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportComparison:  # noqa: F405
        """Get a comparison between two Report Results

        Get a comparison of two Report Results for a Report Specification

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param var_from: Base Report Result UUID to compare
            Format: `uuid`
            Example: `'var_from_example'`
        :param to: Report Result UUID to compare to
            Format: `uuid`
            Example: `'to_example'`
        :param decimal_places: (optional) Number of decimal places to use for comparison (between 0 and 10 inclusive)
            Example: `56`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {
            "from": var_from,
            "to": to,
        }
        if decimal_places is not None:
            _query_params["decimal_places"] = decimal_places

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ReportComparison,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}/comparisons",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_report_results_list(
        self,
        rs_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        offset: int | None = None,
        limit: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PaginatedResponseReportResultSummary:  # noqa: F405
        """List Report Results for a Report Specification

        Gets all Report Results for a Report Specification identified by `rs_id`. The list is ordered from the most recently created to the least recently created Report Result.  This endpoint is paginated, therefore by default this lists, at most, the first 50 items. To get other items, use the `offset` and `limit` query parameters to select the desired part of the list. An `offset` beyond the total number of Report Results for the Report Specification will result in an empty `results` list. The `limit` must be an integer from 1 to 100. The response includes `total`, which is the total number of items within the list.  All workspace roles can use this endpoint.

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param offset: (optional) Index of the first item to return.
            Example: `0`
        :param limit: (optional) Maximum number of items to return from the list. Must be greater than 0 and less than or equal to 100.
            Example: `50`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if offset is not None:
            _query_params["offset"] = offset
        if limit is not None:
            _query_params["limit"] = limit

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": PaginatedResponseReportResultSummary,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}/results",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_report_specification(
        self,
        rs_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportSpecificationWithLastRunInfo:  # noqa: F405
        """Get a Report Specification

        Get an individual Report Specification

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ReportSpecificationWithLastRunInfo,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_block_model_report_specifications(
        self,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        offset: int | None = None,
        limit: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PaginatedResponseWithUnitsReportSpecificationWithLastRunInfo:  # noqa: F405
        """List Report Specifications for a block model

        Gets all Report Specifications for the block model `bm_id`.  The list is ordered alphabetically in ascending order.  This endpoint is paginated, therefore by default this lists, at most, the first 50 items. To get other items, use the `offset` and `limit` query parameters to select the desired part of the list. An `offset` beyond the total number of Report Specifications for the block model will result in an empty `results` list. The `limit` must be an integer from 1 to 100. The response includes `total`, which is the total number of items within the list.  All workspace roles can use this endpoint.

        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param offset: (optional) Index of the first item to return.
            Example: `0`
        :param limit: (optional) Maximum number of items to return from the list. Must be greater than 0 and less than or equal to 100.
            Example: `50`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if offset is not None:
            _query_params["offset"] = offset
        if limit is not None:
            _query_params["limit"] = limit

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": PaginatedResponseWithUnitsReportSpecificationWithLastRunInfo,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def request_report_comparison(
        self,
        rs_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        report_comparison_spec: ReportComparisonSpec,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportComparisonRequestResult:  # noqa: F405
        """Run a Report Comparison Job

        Request a Report Comparison between two versions.  This will either create a new comparison job, or return Report Result IDs for existing report results if both are already available.

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param report_comparison_spec:
            Example: `endpoints.ReportComparisonSpec()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ReportComparisonRequestResult,  # noqa: F405
            "201": ReportComparisonRequestResult,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}/comparisons",
            path_params=_path_params,
            header_params=_header_params,
            body=report_comparison_spec,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def run_reporting_job(
        self,
        rs_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        reporting_job_spec: ReportingJobSpec,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportingJobResult:  # noqa: F405
        """Run a Reporting Job

        Run a new Reporting Job.  The Reporting Job will generate a Report Result for the provided block model version using the latest revision of the Report Specification.

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param reporting_job_spec:
            Example: `endpoints.ReportingJobSpec()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ReportingJobResult,  # noqa: F405
            "201": ReportingJobResult,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}/jobs",
            path_params=_path_params,
            header_params=_header_params,
            body=reporting_job_spec,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def update_report_specification(
        self,
        rs_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        create_report_specification: CreateReportSpecification,  # noqa: F405
        run_now: bool | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ReportSpecificationWithJobUrl:  # noqa: F405
        """Update a Report Specification

        Update a Report Specification for the block model.  Changes to any properties other than `name`, `description`, and `autorun` will incrememnt the `revision` number on the Report Specification, as results based on older revisions may no longer be comparable.

        :param rs_id: ID of the Report Specification that this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e478e8de-f6ba-4920-8c6e-0882e90f00e2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param create_report_specification:
            Example: `endpoints.CreateReportSpecification()`
        :param run_now: (optional) Whether to trigger a Reporting Job immediately after the update of the Report Specification
            Example: `False`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "rs_id": rs_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if run_now is not None:
            _query_params["run_now"] = run_now

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ReportSpecificationWithJobUrl,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.PATCH,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/report-specifications/{rs_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=create_report_specification,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from evo.common.exceptions import EvoClientException

from .endpoints.models import JobErrorPayload


class JobFailedException(EvoClientException):
    """Exception raised when a job fails.

    :param job_id: ID of the job that failed.
    :param error_payload: The error payload of the failed job.
    """

    def __init__(self, job_id: UUID, error_payload: JobErrorPayload):
        self.job_id = job_id
        self.error_payload = error_payload
        super().__init__(f"Job {job_id} failed: {self.error_payload.title}")


class UnknownJobPayload(EvoClientException):
    def __init__(self, job_id: UUID, error_message: str):
        self.job_id = job_id
        super().__init__(error_message)


class CacheNotConfiguredException(EvoClientException):
    """Exception raised when the cache is not configured."""


class MissingColumnInTable(EvoClientException):
    """Exception raised when the provided table is missing some columns."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message
